package tcc.haruki.schnorr;

import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import tcc.haruki.domain.XTRElem;
import tcc.haruki.domain.xtr.XTRKey;
import tcc.haruki.domain.xtr.XTRParameters;
import tcc.haruki.domain.xtr.XTRPrivateKey;
import tcc.haruki.domain.xtr.XTRPublicKey;
import tcc.haruki.exceptions.XTRInvalidKeyException;
import tcc.haruki.exceptions.XTRInvalidParametersException;
import tcc.haruki.exceptions.XTRNullKeyException;
import tcc.haruki.exceptions.XTRNullParametersException;
import tcc.haruki.schnorr.exceptions.SchnorrException;
import tcc.haruki.services.XTROperationService;
import tcc.haruki.services.impl.XTRDefaultOperationService;

public class SchnorrXTRSignature {
	
	private XTRPrivateKey privateKey;
	private XTRPublicKey publicKey;
	
	private byte[] msg;
	
	private MessageDigest md;
	
	private KeyStatus status;
	
	private XTROperationService service;
	
	private SchnorrXTRSignature() throws NoSuchAlgorithmException {
		privateKey = null;
		publicKey = null;
		msg = null;
		md = MessageDigest.getInstance("SHA1");
	}
	
	public static SchnorrXTRSignature initialize(XTRPrivateKey pkey) throws XTRNullKeyException, NoSuchAlgorithmException {
		SchnorrXTRSignature signature = new SchnorrXTRSignature();
		if (pkey == null) {
			throw new XTRNullKeyException();
		}
		
		signature.privateKey = pkey;
		signature.status = KeyStatus.INITIALIZED;
		
		return signature;
	}
	
	public static SchnorrXTRSignature initialize(XTRPublicKey pkey) throws XTRNullKeyException, NoSuchAlgorithmException {
		SchnorrXTRSignature signature = new SchnorrXTRSignature();
		if (pkey == null) {
			throw new XTRNullKeyException();
		}
		
		signature.publicKey = pkey;
		signature.status = KeyStatus.INITIALIZED;
		
		return signature;
	}
	
	public static SchnorrXTRSignature initialize(XTRKey key) throws XTRNullKeyException, XTRInvalidKeyException, NoSuchAlgorithmException, XTRInvalidParametersException {
		SchnorrXTRSignature signature = new SchnorrXTRSignature();
		if (key == null) {
			throw new XTRNullKeyException();
		}
		if (!key.isValid()) {
			throw new XTRInvalidKeyException();
		}
		
		signature.privateKey = key.getPrivateKey();
		signature.publicKey = key.getPublicKey();
		signature.status = KeyStatus.INITIALIZED;

		return signature;
	}
	
	public void update(byte[] msg) throws SchnorrException, NullPointerException  {
		if (!(this.status.equals(KeyStatus.INITIALIZED) || this.status.equals(KeyStatus.UPDATED))) {
			throw new SchnorrException("Atualização de mensagem não realizada - " +
					"Operação não permitida após iniciar-se uma criação ou verificação de assinatura");
		}
		
		this.msg = msg;
		this.status = KeyStatus.UPDATED;
	}
	
	public void startSigning() throws SchnorrException, NullPointerException {
		if (!(this.status.equals(KeyStatus.INITIALIZED) || this.status.equals(KeyStatus.UPDATED))) {
			throw new SchnorrException("Início de criação de assinatura não realizada - " +
					"Operação não permitida após iniciar-se uma criação ou verificação de assinatura");
		}
		if (this.status.equals(this.status.equals(KeyStatus.INITIALIZED))) {
			throw new SchnorrException("Início de criação de assinatura não realizada - " +
					"Operação não permitida sem atualizar a mensagem a ser assinada");
		}
		this.status = KeyStatus.SIGNING;
	}
	
	public void startVerifying() throws SchnorrException, NullPointerException {
		if (!(this.status.equals(KeyStatus.INITIALIZED) || this.status.equals(KeyStatus.UPDATED))) {
			throw new SchnorrException("Início de verificação de assinatura não realizada - " +
					"Operação não permitida após iniciar-se uma criação ou verificação de assinatura");
		}
		if (this.status.equals(this.status.equals(KeyStatus.INITIALIZED))) {
			throw new SchnorrException("Início de verificação de assinatura não realizada - " +
					"Operação não permitida sem atualizar a mensagem a ser assinada");
		}
		this.status = KeyStatus.VERIFYING;
	}
	
	public byte[] sign() throws SchnorrException, NullPointerException, NoSuchAlgorithmException, XTRInvalidParametersException, XTRNullParametersException {
		if (! (this.status == KeyStatus.SIGNING)) {
			throw new SchnorrException("Criação de assinatura não realizada - " +
					"Operação não permitida sem atualizar inicializer o processo usando o método startSigning");
		}
		XTRParameters params = XTRParameters.makeParametersWith(privateKey.getP(), privateKey.getQ());
		service = new XTRDefaultOperationService(params);
		BigInteger x = service.random1AtehQ();
		XTRElem r = service.exponenciacaoSimples(privateKey.getTrg(), x);
		
		BigInteger e = hash(concatenate(msg, r));
		
		BigInteger s = x.subtract(privateKey.getK().multiply(e)).mod(privateKey.getQ());
		
		return getByteRepresentation(e, s);
	}
	
	public boolean verify(byte[] sign) throws SchnorrException, NullPointerException, NoSuchAlgorithmException, XTRInvalidParametersException, XTRNullParametersException {
		if (! (this.status == KeyStatus.VERIFYING)) {
			throw new SchnorrException("Verificação de assinatura não realizada - " +
					"Operação não permitida sem atualizar inicializer o processo usando o método startVerifying");
		}
		
		XTRParameters params = XTRParameters.makeParametersWith(publicKey.getP(), publicKey.getQ());
		service = new XTRDefaultOperationService(params);
		
		BigInteger[] vet = getFromByteRepresentation(sign);
		
		BigInteger e = vet[0];
		BigInteger s = vet[1];
		
		XTRElem rv = service.exponenciacaoDupla(publicKey.getTrg(), publicKey.getH(), s, e);
		
		BigInteger ev = hash(concatenate(msg, rv));
		
		if (ev.equals(e)) {
			return true;
		}
		return false;
	}
	
	private BigInteger[] getFromByteRepresentation(byte[] barray) {
		int size1;
		int size2;
		BigInteger retval[] = new BigInteger[2];
		
		byte[] size1b = new byte[4];
		byte[] size2b = new byte[4];
		
		for (int i = 0; i < 4; i++) {
			size1b[i] = barray[i];
		}
		for (int i = 0; i < 4; i++) {
			size2b[i] = barray[i+4];
		}
		
		size1 = new BigInteger(size1b).intValue();
		size2 = new BigInteger(size2b).intValue();
		
		byte[] num1b = new byte[size1];
		byte[] num2b = new byte[size2];
		
		int idx = 8;
		for (int i = 0; i < size1; i++) {
			num1b[i] = barray[idx++];
		}
		for (int i = 0; i < size2; i++) {
			num2b[i] = barray[idx++];
		}
		
		retval[0] = new BigInteger(num1b);
		retval[1] = new BigInteger(num2b);
		
		return retval;
	}
	
	private byte[] getByteRepresentation(BigInteger a, BigInteger b) {
		byte[] rep1 = a.toByteArray();
		byte[] rep2 = b.toByteArray();
		
		int size1 = rep1.length;
		int size2 = rep2.length;
		
		byte[] size1b = new BigInteger(Integer.toString(size1)).toByteArray();
		byte[] size2b = new BigInteger(Integer.toString(size2)).toByteArray();
		
		byte[] ret = new byte[rep1.length + rep2.length + 8];
		
		for (int i = 0; i < 8; i++) {
			ret[i] = 0;
		}
		
		int idx = 3;
		for (int i = size1b.length-1; i >= 0; i--) {
			ret[idx--] = size1b[i];
		}
		
		idx = 7;
		for (int i = size2b.length-1; i >= 0; i--) {
			ret[idx--] = size2b[i];
		}
	
		idx = 8;
		for (int i = 0; i < rep1.length; i++) {
			ret[idx++] = rep1[i];
		}
		for (int i = 0; i < rep2.length; i++) {
			ret[idx++] = rep2[i];
		}
		
		return ret;
	}
	
	private byte[] concatenate (byte[] msg, XTRElem r) {
		byte[] rbyte = r.getByteRepresentation();
		byte[] retval = new byte[msg.length + rbyte.length];
		for (int i = 0; i < msg.length; i++) {
			retval[i] = msg[i]; 
		}
		for (int i = 0; i < rbyte.length; i++) {
			retval[i+msg.length] = rbyte[i];
		}
		return retval;
	}
	
	private BigInteger hash(byte[] m) throws NoSuchAlgorithmException {
		md.update(m);
		byte[] digest = md.digest();
		
		BigInteger a = new BigInteger(1, digest);
		return a;
	}
	
	private enum KeyStatus {
		INITIALIZED,
		UPDATED,
		SIGNING,
		VERIFYING,
		DONE
	}
}
